# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from cherab.core.utility.conversion import PhotonToJ

cimport cython
from cherab.core.math cimport Interpolate1DCubic, Interpolate2DCubic

# todo: clarify variables

cdef class BeamStoppingRate(CoreBeamStoppingRate):
    """
    The beam stopping coefficient interpolation class.

    :param data: A dictionary holding the beam coefficient data.
    :param extrapolate: Set to True to enable extrapolation, False to disable (default).
    """

    @cython.cdivision(True)
    def __init__(self, dict data, bint extrapolate=False):

        self.raw_data = data

        # unpack
        e = data["e"]                          # eV/amu
        n = data["n"]                          # m^-3
        t = data["t"]                          # eV
        sen = data["sen"]                      # m^3/s
        st = data["st"] / data["sref"]         # dimensionless

        # store limits of data
        self.beam_energy_range = e.min(), e.max()
        self.density_range = n.min(), n.max()
        self.temperature_range = t.min(), t.max()

        # interpolate
        self._npl_eb = Interpolate2DCubic(e, n, sen, tolerate_single_value=True, extrapolate=extrapolate, extrapolation_type="quadratic")
        self._tp = Interpolate1DCubic(t, st, tolerate_single_value=True, extrapolate=extrapolate, extrapolation_type="quadratic")

    cpdef double evaluate(self, double energy, double density, double temperature) except? -1e999:
        """
        Interpolates and returns the beam coefficient for the supplied parameters.

        If the requested data is out-of-range then the call with throw a ValueError exception.

        :param energy: Interaction energy in eV/amu.
        :param density: Target electron density in m^-3
        :param temperature: Target temperature in eV.
        :return: The beam stopping coefficient in m^3.s^-1
        """

        cdef double val

        val = self._npl_eb.evaluate(energy, density)
        if val <= 0:
            return 0.0

        val *= self._tp.evaluate(temperature)
        if val <= 0:
            return 0.0

        return val


cdef class BeamPopulationRate(CoreBeamPopulationRate):
    """
    The beam population coefficient interpolation class.

    :param data: A dictionary holding the beam coefficient data.
    :param extrapolate: Set to True to enable extrapolation, False to disable (default).
    """

    @cython.cdivision(True)
    def __init__(self, dict data, bint extrapolate=False):

        self.raw_data = data

        # unpack
        e = data["e"]                          # eV/amu
        n = data["n"]                          # m^-3
        t = data["t"]                          # eV
        sen = data["sen"]                      # dimensionless
        st = data["st"] / data["sref"]         # dimensionless

        # store limits of data
        self.beam_energy_range = e.min(), e.max()
        self.density_range = n.min(), n.max()
        self.temperature_range = t.min(), t.max()

        # interpolate
        self._npl_eb = Interpolate2DCubic(e, n, sen, tolerate_single_value=True, extrapolate=extrapolate, extrapolation_type="quadratic")
        self._tp = Interpolate1DCubic(t, st, tolerate_single_value=True, extrapolate=extrapolate, extrapolation_type="quadratic")

    cpdef double evaluate(self, double energy, double density, double temperature) except? -1e999:
        """
        Interpolates and returns the beam coefficient for the supplied parameters.

        If the requested data is out-of-range then the call with throw a ValueError exception.

        :param energy: Interaction energy in eV/amu.
        :param density: Target electron density in m^-3
        :param temperature: Target temperature in eV.
        :return: The beam population coefficient in dimensionless units.
        """

        cdef double val

        val = self._npl_eb.evaluate(energy, density)
        if val <= 0:
            return 0.0

        val *= self._tp.evaluate(temperature)
        if val <= 0:
            return 0.0

        return val


cdef class BeamEmissionRate(CoreBeamEmissionRate):
    """
    The beam emission coefficient interpolation class.

    :param data: A dictionary holding the beam coefficient data.
    :param wavelength: The natural wavelength of the emission line associated with the rate data in nm.
    :param extrapolate: Set to True to enable extrapolation, False to disable (default).
    """

    @cython.cdivision(True)
    def __init__(self, dict data, double wavelength, bint extrapolate=False):

        self.raw_data = data

        # unpack
        e = data["e"]                                   # eV/amu
        n = data["n"]                                   # m^-3
        t = data["t"]                                   # eV
        sen = PhotonToJ.to(data["sen"], wavelength)     # W.m^3/s
        st = data["st"] / data["sref"]                  # dimensionless

        # store limits of data
        self.beam_energy_range = e.min(), e.max()
        self.density_range = n.min(), n.max()
        self.temperature_range = t.min(), t.max()

        # interpolate
        self._npl_eb = Interpolate2DCubic(e, n, sen, tolerate_single_value=True, extrapolate=extrapolate, extrapolation_type="quadratic")
        self._tp = Interpolate1DCubic(t, st, tolerate_single_value=True, extrapolate=extrapolate, extrapolation_type="quadratic")

    cpdef double evaluate(self, double energy, double density, double temperature) except? -1e999:
        """
        Interpolates and returns the beam coefficient for the supplied parameters.

        If the requested data is out-of-range then the call with throw a ValueError exception.

        :param energy: Interaction energy in eV/amu.
        :param density: Target electron density in m^-3
        :param temperature: Target temperature in eV.
        :return: The beam emission coefficient in m^3.s^-1
        """

        cdef double val

        val = self._npl_eb.evaluate(energy, density)
        if val <= 0:
            return 0.0

        val *= self._tp.evaluate(temperature)
        if val <= 0:
            return 0.0

        return val



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import os
import json
import numpy as np
from cherab.core.atomic import Element
from ..utility import DEFAULT_REPOSITORY_PATH, valid_ionisation

"""
Utilities for managing the local rate repository - beam stopping section.
"""


def add_beam_stopping_rate(beam_species, target_ion, target_ionisation, rate, repository_path=None):
    """
    Adds a single beam stopping/excitation rate to the repository.

    :param beam_species:
    :param target_ion:
    :param target_ionisation:
    :param rate:
    :return:
    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    # sanitise and validate arguments
    if not isinstance(beam_species, Element):
        raise TypeError('The beam_species must be an Element object.')

    if not isinstance(target_ion, Element):
        raise TypeError('The beam_species must be an Element object.')

    if not valid_ionisation(target_ion, target_ionisation):
        raise ValueError('Ionisation level is larger than the number of protons in the target ion.')

    # sanitise and validate rate data
    e = np.array(rate['e'], np.float64)
    n = np.array(rate['n'], np.float64)
    t = np.array(rate['t'], np.float64)
    sen = np.array(rate['sen'], np.float64)
    st = np.array(rate['st'], np.float64)

    if e.ndim != 1:
        raise ValueError('Beam energy array must be a 1D array.')

    if n.ndim != 1:
        raise ValueError('Density array must be a 1D array.')

    if t.ndim != 1:
        raise ValueError('Temperature array must be a 1D array.')

    if (e.shape[0], n.shape[0]) != sen.shape:
        raise ValueError('Beam energy, density and combined rate data arrays have inconsistent sizes.')

    if t.shape != st.shape:
        raise ValueError('Temperature and temperature rate data arrays have inconsistent sizes.')

    # convert numpy arrays to lists for json export
    rate['e'] = e.tolist()
    rate['n'] = n.tolist()
    rate['t'] = t.tolist()
    rate['sen'] = sen.tolist()
    rate['st'] = st.tolist()

    rate['eref'] = float(rate['eref'])
    rate['nref'] = float(rate['nref'])
    rate['tref'] = float(rate['tref'])
    rate['sref'] = float(rate['sref'])

    path = os.path.join(repository_path, 'beam/stopping/{}/{}/{}.json'.format(beam_species.symbol.lower(), target_ion.symbol.lower(), target_ionisation))

    # create directory structure if missing
    directory = os.path.dirname(path)
    if not os.path.isdir(directory):
        os.makedirs(directory)

    # write new data (simply overwrite any existing rate)
    with open(path, 'w') as f:
        json.dump(rate, f, indent=2, sort_keys=True)


def update_beam_stopping_rates(rates, repository_path=None):
    """
    Beam stopping rate file structure

    /beam/stopping/<beam species>/<target ion>/<target_ionisation>.json

    Each json file contains a single rate, so it can simply be replaced.
    """

    for beam_species, target_ions in rates.items():
        for target_ion, target_ionisations in target_ions.items():
            for target_ionisation, rate in target_ionisations.items():
                add_beam_stopping_rate(beam_species, target_ion, target_ionisation, rate, repository_path)


def get_beam_stopping_rate(beam_species, target_ion, target_ionisation, repository_path=None):

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH
    path = os.path.join(repository_path, 'beam/stopping/{}/{}/{}.json'.format(beam_species.symbol.lower(), target_ion.symbol.lower(), target_ionisation))
    try:
        with open(path, 'r') as f:
            rate = json.load(f)
    except FileNotFoundError:
        raise RuntimeError('Requested beam stopping rate (beam species={}, target ion={}, target ionisation={})'
                           ' is not available.'.format(beam_species.symbol, target_ion.symbol, target_ionisation))

    # convert lists to numpy arrays
    rate['e'] = np.array(rate['e'], np.float64)
    rate['n'] = np.array(rate['n'], np.float64)
    rate['t'] = np.array(rate['t'], np.float64)
    rate['sen'] = np.array(rate['sen'], np.float64)
    rate['st'] = np.array(rate['st'], np.float64)

    return rate



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from cherab.core cimport ImpactExcitationRate as CoreImpactExcitationRate
from cherab.core cimport RecombinationRate as CoreRecombinationRate
from cherab.core cimport ThermalCXRate as CoreThermalCXRate
from cherab.core.math cimport Interpolate2DCubic


cdef class ImpactExcitationRate(CoreImpactExcitationRate):

    cdef:
        readonly dict raw_data
        readonly double wavelength
        readonly tuple density_range, temperature_range
        Interpolate2DCubic _rate


cdef class RecombinationRate(CoreRecombinationRate):

    cdef:
        readonly dict raw_data
        readonly double wavelength
        readonly tuple density_range, temperature_range
        Interpolate2DCubic _rate


# cdef class CXThermalRate(CoreCXThermalRate):
#     pass



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from cherab.core.utility import RecursiveDict
from cherab.core.atomic.elements import *
from cherab.openadas.install import install_files
from cherab.openadas import repository


def populate(download=True, repository_path=None, adas_path=None):
    """
    Populates the OpenADAS repository with a typical set of rates and wavelengths.

    If an ADAS file is not note found an attempt will be made to download the
    file from the OpenADAS website. This behaviour can be disabled by setting
    the download argument to False.

    :param download: Attempt to download the ADAS files if missing (default=True).
    :param repository_path: Alternate path for the OpenADAS repository (default=None).
    :param adas_path: Alternate path in which to search for ADAS files (default=None) .
    """

    # install a common selection of open adas files
    rates = {
        # 'adf11plt': (
        #     (hydrogen, "adf11/plt12/plt12_h.dat"),
        #     (helium, "adf11/plt96/plt96_he.dat"),
        #     (lithium, "adf11/plt96/plt96_li.dat"),
        #     (beryllium, "adf11/plt96/plt96_be.dat"),
        #     (carbon, "adf11/plt96/plt96_c.dat"),
        #     (nitrogen, "adf11/plt96/plt96_n.dat"),
        #     (oxygen,  "adf11/plt96/plt96_o.dat"),
        #     (neon, "adf11/plt96/plt96_ne.dat"),
        #     (argon, "adf11/plt40/plt40_ar.dat"),
        #     (krypton, "adf11/plt89/plt89_kr.dat")
        # ),
        # 'adf11prb': (
        #     (hydrogen, "adf11/prb12/prb12_h.dat"),
        #     (helium, "adf11/prb96/prb96_he.dat"),
        #     (lithium, "adf11/prb96/prb96_li.dat"),
        #     (beryllium, "adf11/prb96/prb96_be.dat"),
        #     (carbon, "adf11/prb96/prb96_c.dat"),
        #     (nitrogen, "adf11/prb96/prb96_n.dat"),
        #     (oxygen, "adf11/prb96/prb96_o.dat"),
        #     (neon, "adf11/prb96/prb96_ne.dat"),
        #     (argon, "adf11/prb89/prb89_ar.dat"),
        #     (krypton, "adf11/prb89/prb89_kr.dat")
        # ),
        'adf12': (
            # (donor, receiver, ionisation, donor_metastable, rate file)
            (hydrogen, 1, hydrogen,  1, 'adf12/qef93#h/qef93#h_h1.dat'),
            (hydrogen, 1, helium,    2, "adf12/qef93#h/qef93#h_he2.dat"),
            (hydrogen, 2, helium,    2, "adf12/qef97#h/qef97#h_en2_kvi#he2.dat"),
            (hydrogen, 1, beryllium, 4, "adf12/qef93#h/qef93#h_be4.dat"),
            (hydrogen, 2, beryllium, 4, "adf12/qef97#h/qef97#h_en2_kvi#be4.dat"),
            (hydrogen, 1, boron,     5, "adf12/qef93#h/qef93#h_b5.dat"),
            (hydrogen, 2, boron,     5, "adf12/qef97#h/qef97#h_en2_kvi#b5.dat"),
            (hydrogen, 1, carbon,    6, "adf12/qef93#h/qef93#h_c6.dat"),
            (hydrogen, 2, carbon,    6, "adf12/qef97#h/qef97#h_en2_kvi#c6.dat"),
            (hydrogen, 1, neon,      10, "adf12/qef93#h/qef93#h_ne10.dat"),
            (hydrogen, 2, neon,      10, "adf12/qef97#h/qef97#h_en2_kvi#ne10.dat")
        ),
        'adf15': (
            (hydrogen,  0, 'adf15/pec12#h/pec12#h_pju#h0.dat'),
            (helium,    0, 'adf15/pec96#he/pec96#he_pju#he0.dat'),
            (helium,    1, 'adf15/pec96#he/pec96#he_pju#he1.dat'),
            (beryllium, 0, 'adf15/pec96#be/pec96#be_pju#be0.dat'),
            (beryllium, 1, 'adf15/pec96#be/pec96#be_pju#be1.dat'),
            (beryllium, 2, 'adf15/pec96#be/pec96#be_pju#be2.dat'),
            (beryllium, 3, 'adf15/pec96#be/pec96#be_pju#be3.dat'),
            (carbon,    0, 'adf15/pec96#c/pec96#c_vsu#c0.dat'),
            (carbon,    1, 'adf15/pec96#c/pec96#c_vsu#c1.dat'),
            (carbon,    2, 'adf15/pec96#c/pec96#c_vsu#c2.dat'),
            # (neon,      0, 'adf15/pec96#ne/pec96#ne_pju#ne0.dat'),     #TODO: OPENADAS DATA CORRUPT
            # (neon,      1, 'adf15/pec96#ne/pec96#ne_pju#ne1.dat'),     #TODO: OPENADAS DATA CORRUPT
            (nitrogen,  0, 'adf15/pec96#n/pec96#n_vsu#n0.dat'),
            (nitrogen,  1, 'adf15/pec96#n/pec96#n_vsu#n1.dat'),
            # (nitrogen,  2, 'adf15/pec96#n/pec96#n_vsu#n2.dat'),    #TODO: OPENADAS DATA CORRUPT
        ),
        'adf21': (
            # (beam_species, target_ion, target_ionisation, rate file)
            (hydrogen, hydrogen,  1,  "adf21/bms97#h/bms97#h_h1.dat"),
            (hydrogen, helium,    2,  "adf21/bms97#h/bms97#h_he2.dat"),
            (hydrogen, lithium,   3,  "adf21/bms97#h/bms97#h_li3.dat"),
            (hydrogen, beryllium, 4,  "adf21/bms97#h/bms97#h_be4.dat"),
            (hydrogen, boron,     5,  "adf21/bms97#h/bms97#h_b5.dat"),
            (hydrogen, carbon,    6,  "adf21/bms97#h/bms97#h_c6.dat"),
            (hydrogen, nitrogen,  7,  "adf21/bms97#h/bms97#h_n7.dat"),
            (hydrogen, oxygen,    8,  "adf21/bms97#h/bms97#h_o8.dat"),
            (hydrogen, fluorine,  9,  "adf21/bms97#h/bms97#h_f9.dat"),
            (hydrogen, neon,      10, "adf21/bms97#h/bms97#h_ne10.dat"),
        ),
        'adf22bmp': (
            # (beam species, beam metastable, target ion, target ionisation, rate file)
            (hydrogen, 2, hydrogen,  1,  "adf22/bmp97#h/bmp97#h_2_h1.dat"),
            (hydrogen, 3, hydrogen,  1,  "adf22/bmp97#h/bmp97#h_3_h1.dat"),
            (hydrogen, 4, hydrogen,  1,  "adf22/bmp97#h/bmp97#h_4_h1.dat"),
            (hydrogen, 2, helium,    2,  "adf22/bmp97#h/bmp97#h_2_he2.dat"),
            (hydrogen, 3, helium,    2,  "adf22/bmp97#h/bmp97#h_3_he2.dat"),
            (hydrogen, 4, helium,    2,  "adf22/bmp97#h/bmp97#h_4_he2.dat"),
            (hydrogen, 2, lithium,   3,  "adf22/bmp97#h/bmp97#h_2_li3.dat"),
            (hydrogen, 3, lithium,   3,  "adf22/bmp97#h/bmp97#h_3_li3.dat"),
            (hydrogen, 4, lithium,   3,  "adf22/bmp97#h/bmp97#h_4_li3.dat"),
            (hydrogen, 2, beryllium, 4,  "adf22/bmp97#h/bmp97#h_2_be4.dat"),
            (hydrogen, 3, beryllium, 4,  "adf22/bmp97#h/bmp97#h_3_be4.dat"),
            (hydrogen, 4, beryllium, 4,  "adf22/bmp97#h/bmp97#h_4_be4.dat"),
            (hydrogen, 2, boron,     5,  "adf22/bmp97#h/bmp97#h_2_b5.dat"),
            (hydrogen, 3, boron,     5,  "adf22/bmp97#h/bmp97#h_3_b5.dat"),
            (hydrogen, 4, boron,     5,  "adf22/bmp97#h/bmp97#h_4_b5.dat"),
            (hydrogen, 2, carbon,    6,  "adf22/bmp97#h/bmp97#h_2_c6.dat"),
            (hydrogen, 3, carbon,    6,  "adf22/bmp97#h/bmp97#h_3_c6.dat"),
            (hydrogen, 4, carbon,    6,  "adf22/bmp97#h/bmp97#h_4_c6.dat"),
            (hydrogen, 2, nitrogen,  7,  "adf22/bmp97#h/bmp97#h_2_n7.dat"),
            (hydrogen, 3, nitrogen,  7,  "adf22/bmp97#h/bmp97#h_3_n7.dat"),
            (hydrogen, 4, nitrogen,  7,  "adf22/bmp97#h/bmp97#h_4_n7.dat"),
            (hydrogen, 2, oxygen,    8,  "adf22/bmp97#h/bmp97#h_2_o8.dat"),
            (hydrogen, 3, oxygen,    8,  "adf22/bmp97#h/bmp97#h_3_o8.dat"),
            (hydrogen, 4, oxygen,    8,  "adf22/bmp97#h/bmp97#h_4_o8.dat"),
            (hydrogen, 2, fluorine,  9,  "adf22/bmp97#h/bmp97#h_2_f9.dat"),
            (hydrogen, 3, fluorine,  9,  "adf22/bmp97#h/bmp97#h_3_f9.dat"),
            (hydrogen, 4, fluorine,  9,  "adf22/bmp97#h/bmp97#h_4_f9.dat"),
            (hydrogen, 2, neon,      10, "adf22/bmp97#h/bmp97#h_2_ne10.dat"),
            (hydrogen, 3, neon,      10, "adf22/bmp97#h/bmp97#h_3_ne10.dat"),
            (hydrogen, 4, neon,      10, "adf22/bmp97#h/bmp97#h_4_ne10.dat"),
        ),
        'adf22bme': (
            # (beam species, target_ion, target_ionisation, (initial_level, final_level), rate file)
            (hydrogen, hydrogen,  1,  (3, 2), "adf22/bme10#h/bme10#h_h1.dat"),
            (hydrogen, helium,    2,  (3, 2), "adf22/bme97#h/bme97#h_he2.dat"),
            (hydrogen, lithium,   3,  (3, 2), "adf22/bme97#h/bme97#h_li3.dat"),
            (hydrogen, beryllium, 4,  (3, 2), "adf22/bme97#h/bme97#h_be4.dat"),
            (hydrogen, boron,     5,  (3, 2), "adf22/bme97#h/bme97#h_b5.dat"),
            (hydrogen, carbon,    6,  (3, 2), "adf22/bme97#h/bme97#h_c6.dat"),
            (hydrogen, nitrogen,  7,  (3, 2), "adf22/bme97#h/bme97#h_n7.dat"),
            (hydrogen, oxygen,    8,  (3, 2), "adf22/bme97#h/bme97#h_o8.dat"),
            (hydrogen, fluorine,  9,  (3, 2), "adf22/bme97#h/bme97#h_f9.dat"),
            (hydrogen, neon,      10, (3, 2), "adf22/bme97#h/bme97#h_ne10.dat"),
            (hydrogen, argon,     18, (3, 2), "adf22/bme99#h/bme99#h_ar18.dat"),
        )
    }

    # add common wavelengths to the repository
    wavelengths = RecursiveDict()

    # H0
    wavelengths[hydrogen][0] = {
        (2, 1): 121.52,
        (3, 1): 102.53,
        (3, 2): 656.19,
        (4, 1): 97.21,
        (4, 2): 486.06,
        (4, 3): 1874.82,
        (5, 1): 94.93,
        (5, 2): 433.99,
        (5, 3): 1281.61,
        (5, 4): 4050.53,
        (6, 1): 93.74,
        (6, 2): 410.12,
        (6, 3): 1093.64,
        (6, 4): 2624.75,
        (6, 5): 7456.66,
        (7, 1): 93.04,
        (7, 2): 396.95,
        (7, 3): 1004.79,
        (7, 4): 2165.19,
        (7, 5): 4651.78,
        (7, 6): 12366.59,
        (8, 1): 92.58,
        (8, 2): 388.85,
        (8, 3): 954.45,
        (8, 4): 1944.26,
        (8, 5): 3738.95,
        (8, 6): 7499.27,
        (8, 7): 19053.71,
        (9, 1): 92.28,
        (9, 2): 383.49,
        (9, 3): 922.76,
        (9, 4): 1817.13,
        (9, 5): 3295.58,
        (9, 6): 5905.68,
        (9, 7): 11303.84,
        (9, 8): 27791.42,
        (10, 1): 92.06,
        (10, 2): 379.74,
        (10, 3): 901.35,
        (10, 4): 1735.94,
        (10, 5): 3037.9,
        (10, 6): 5126.46,
        (10, 7): 8756.3,
        (10, 8): 16202.13,
        (10, 9): 38853.14,
        (11, 1): 91.9,
        (11, 2): 377.01,
        (11, 3): 886.14,
        (11, 4): 1680.39,
        (11, 5): 2871.76,
        (11, 6): 4670.5,
        (11, 7): 7504.88,
        (11, 8): 12381.84,
        (11, 9): 22330.84,
        (11, 10): 52512.27,
        (12, 1): 91.77,
        (12, 2): 374.96,
        (12, 3): 874.92,
        (12, 4): 1640.47,
        (12, 5): 2757.09,
        (12, 6): 4374.58,
        (12, 7): 6769.08,
        (12, 8): 10498.98,
        (12, 9): 16873.36,
        (12, 10): 29826.65,
        (12, 11): 69042.22,
    }

    # todo: temporary - add more accurate values
    wavelengths[deuterium][0] = wavelengths[hydrogen][0]
    wavelengths[tritium][0] = wavelengths[hydrogen][0]

    # He1+
    wavelengths[helium][1] = {
        (2, 1): 30.378,     # 2p -> 1s
        (3, 1): 25.632,     # 3p -> 1s
        (3, 2): 164.04,     # 3d -> 2p
        (4, 2): 121.51,     # 4d -> 2p
        (4, 3): 468.71,     # 4f -> 3d
        (5, 3): 320.28,     # 5f -> 3d
        (5, 4): 1012.65,    # 5g -> 4f
        (6, 4): 656.20,     # 6g -> 4f
        (6, 5): 1864.20,    # 6h -> 5g
        (7, 5): 1162.53,    # from ADAS comment, unknown source
        (7, 6): 3090.55     # from ADAS comment, unknown source
    }

    # Be3+
    wavelengths[beryllium][3] = {
        (3, 1): 6.4065,     # 3p -> 1s
        (3, 2): 41.002,     # 3d -> 2p
        (4, 2): 30.373,     # 4d -> 2p
        (4, 3): 117.16,     # 4f -> 3d
        (5, 3): 80.092,     # 5f -> 3d
        (5, 4): 253.14,     # 5g -> 4f
        (6, 4): 164.03,     # 6g -> 4f
        (6, 5): 466.01,     # 6h -> 5g
        (7, 5): 290.62,     # from ADAS comment, unknown source
        (7, 6): 772.62,     # from ADAS comment, unknown source
        (8, 6): 468.53,     # from ADAS comment, unknown source
        (8, 7): 1190.42     # from ADAS comment, unknown source
    }

    # B4+
    wavelengths[boron][4] = {
        (3, 1): 4.0996,     # 3p -> 1s
        (3, 2): 26.238,     # 3d -> 2p
        (4, 2): 19.437,     # 4d -> 2p
        (4, 3): 74.980,     # 4f -> 3d
        (5, 3): 51.257,     # 5f -> 3d
        (5, 4): 162.00,     # 5g -> 4f
        (6, 4): 104.98,     # 6g -> 4f
        (6, 5): 298.24,     # 6h -> 5g
        (7, 5): 186.05,     # 7h -> 5g
        (7, 6): 494.48,     # 7i -> 6h
        (8, 6): 299.86,     # 8i -> 6h
        (8, 7): 761.87,     # 8k -> 7i
        (9, 7): 451.99,     # 9k -> 7i
        (9, 8): 1111.25     # from ADAS comment, unknown source
    }

    # C5+
    wavelengths[carbon][5] = {
        (4, 2): 13.496,     # 4d -> 2p
        (4, 3): 52.067,     # 4f -> 3d
        (5, 3): 35.594,     # 5f -> 3d
        (5, 4): 112.50,     # 5g -> 4f
        (6, 4): 72.900,     # 6g -> 4f
        (6, 5): 207.11,     # 6h -> 5g
        (7, 5): 129.20,     # from ADAS comment, unknown source
        (7, 6): 343.38,     # from ADAS comment, unknown source
        (8, 6): 208.23,     # from ADAS comment, unknown source
        (8, 7): 529.07,     # from ADAS comment, unknown source
        (9, 7): 313.87,     # from ADAS comment, unknown source
        (9, 8): 771.69,     # from ADAS comment, unknown source
        (10, 8): 449.89,    # from ADAS comment, unknown source
        (10, 9): 1078.86    # from ADAS comment, unknown source
    }

    # Ne9+
    wavelengths[neon][9] = {
        (6, 5): 74.54,      # from ADAS comment, unknown source
        (7, 6): 123.64,     # from ADAS comment, unknown source
        (8, 7): 190.50,     # from ADAS comment, unknown source
        (9, 8): 277.79,     # from ADAS comment, unknown source
        (10, 9): 388.37,    # from ADAS comment, unknown source
        (11, 10): 524.92,   # from ADAS comment, unknown source
        (12, 11): 690.16,   # from ADAS comment, unknown source
        (13, 12): 886.83,   # from ADAS comment, unknown source
        (6, 4): 26.24,      # from ADAS comment, unknown source
        (7, 5): 46.51,      # from ADAS comment, unknown source
        (8, 6): 74.98,      # from ADAS comment, unknown source
        (9, 7): 113.02,     # from ADAS comment, unknown source
        (10, 8): 162.00,    # from ADAS comment, unknown source
        (11, 9): 223.22,    # from ADAS comment, unknown source
        (12, 10): 298.15,   # from ADAS comment, unknown source
        (13, 11): 388.12    # from ADAS comment, unknown source
    }

    install_files(rates, download=download, repository_path=repository_path, adas_path=adas_path)
    repository.update_wavelengths(wavelengths, repository_path=repository_path)



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import os
import json
import numpy as np
from cherab.core.utility import RecursiveDict
from cherab.core.atomic import Element
from ..utility import DEFAULT_REPOSITORY_PATH, valid_ionisation, encode_transition

"""
Utilities for managing the local rate repository - PEC section.
"""


def add_beam_emission_rate(beam_species, target_ion, target_ionisation, transition, rate, repository_path=None):
    """
    Adds a single beam emission rate to the repository.

    If adding multiple rate, consider using the update_beam_emission_rates()
    function instead. The update function avoid repeatedly opening and closing
    the rate files.

    :param repository_path:
    :return:
    """

    update_beam_emission_rates({
        beam_species: {
            target_ion: {
                target_ionisation: {
                    transition: rate
                }
            }
        }
    }, repository_path)


def update_beam_emission_rates(rates, repository_path=None):
    """
    Beam emission rate file structure

    /beam/emission/<beam species>/<target ion>/<target_ionisation>.json

    File contains multiple rates, indexed by transition.
    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    for beam_species, target_ions in rates.items():
        for target_ion, target_ionisations in target_ions.items():
            for target_ionisation, transitions in target_ionisations.items():

                # sanitise and validate arguments
                if not isinstance(beam_species, Element):
                    raise TypeError('The beam_species must be an Element object.')

                if not isinstance(target_ion, Element):
                    raise TypeError('The beam_species must be an Element object.')

                if not valid_ionisation(target_ion, target_ionisation):
                    raise ValueError('Ionisation level is larger than the number of protons in the target ion.')

                path = os.path.join(repository_path, 'beam/emission/{}/{}/{}.json'.format(beam_species.symbol.lower(), target_ion.symbol.lower(), target_ionisation))

                # read in any existing rates
                try:
                    with open(path, 'r') as f:
                        content = RecursiveDict.from_dict(json.load(f))
                except FileNotFoundError:
                    content = RecursiveDict()

                # add/replace data for a transition
                for transition in transitions:
                    key = encode_transition(transition)
                    rate = rates[beam_species][target_ion][target_ionisation][transition]

                    # sanitise and validate rate data
                    e = np.array(rate['e'], np.float64)
                    n = np.array(rate['n'], np.float64)
                    t = np.array(rate['t'], np.float64)
                    sen = np.array(rate['sen'], np.float64)
                    st = np.array(rate['st'], np.float64)

                    if e.ndim != 1:
                        raise ValueError('Beam energy array must be a 1D array.')

                    if n.ndim != 1:
                        raise ValueError('Density array must be a 1D array.')

                    if t.ndim != 1:
                        raise ValueError('Temperature array must be a 1D array.')

                    if (e.shape[0], n.shape[0]) != sen.shape:
                        raise ValueError('Beam energy, density and combined rate data arrays have inconsistent sizes.')

                    if t.shape != st.shape:
                        raise ValueError('Temperature and temperature rate data arrays have inconsistent sizes.')

                    # update file content with new rate
                    content[key] = {
                        'e': e.tolist(),
                        'n': n.tolist(),
                        't': t.tolist(),
                        'sen': sen.tolist(),
                        'st': st.tolist(),
                        'eref': float(rate['eref']),
                        'nref': float(rate['nref']),
                        'tref': float(rate['tref']),
                        'sref': float(rate['sref'])
                    }

                # create directory structure if missing
                directory = os.path.dirname(path)
                if not os.path.isdir(directory):
                    os.makedirs(directory)

                # write new data
                with open(path, 'w') as f:
                    json.dump(content, f, indent=2, sort_keys=True)


def get_beam_emission_rate(beam_species, target_ion, target_ionisation, transition, repository_path=None):

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH
    path = os.path.join(repository_path, 'beam/stopping/{}/{}/{}.json'.format(beam_species.symbol.lower(), target_ion.symbol.lower(), target_ionisation))
    try:
        with open(path, 'r') as f:
            content = json.load(f)
        rate = content[encode_transition(transition)]
    except (FileNotFoundError, KeyError):
        raise RuntimeError('Requested beam emission rate (beam species={}, target ion={}, target ionisation={}, transition={})'
                           ' is not available.'.format(beam_species.symbol, target_ion.symbol, target_ionisation, transition))

    # convert lists to numpy arrays
    rate['e'] = np.array(rate['e'], np.float64)
    rate['n'] = np.array(rate['n'], np.float64)
    rate['t'] = np.array(rate['t'], np.float64)
    rate['sen'] = np.array(rate['sen'], np.float64)
    rate['st'] = np.array(rate['st'], np.float64)

    return rate


# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import re
import numpy as np
from cherab.core.atomic import hydrogen, Element
from cherab.core.utility import RecursiveDict
from cherab.core.utility.conversion import Cm3ToM3, PerCm3ToPerM3


_L_LOOKUP = {
    0: 'S',
    1: 'P',
    2: 'D',
    3: 'F',
    4: 'G',
    5: 'H',
    6: 'I',
    7: 'K',
    8: 'L',
    9: 'M',
    10: 'N',
    11: 'O',
    12: 'Q',
    13: 'R',
}


def parse_adf15(element, ionisation, adf_file_path):
    """
    Opens and parses ADAS ADF15 data files.

    :param element: Element described by ADF file.
    :param ionisation: Ionisation described by ADF file.
    :param adf_file_path: Path to ADF15 file from ADAS root.
    :return: Dictionary containing rates.
    """

    if not isinstance(element, Element):
        raise TypeError('The element must be an Element object.')

    ionisation = int(ionisation)

    with open(adf_file_path, "r") as file:

        # for check header line
        header = file.readline()
        if not re.match('^\s*(\d*) {4}/(.*)/?\s*$', header):
            raise ValueError('The specified path does not point to a valid ADF15 file.')

        # scrape transition information and wavelength
        config = _scrape_metadata(file, element, ionisation)

        # process rate data
        rates = RecursiveDict()
        for cls in ('excitation', 'recombination', 'thermalcx'):
            for element, ionisations in config[cls].items():
                for ionisation, transitions in ionisations.items():
                    for transition in transitions.keys():
                        block_num = config[cls][element][ionisation][transition]
                        rates[cls][element][ionisation][transition] = _extract_rate(file, block_num)

    wavelengths = config['wavelength']
    return rates, wavelengths


def _scrape_metadata(file, element, ionisation):
    """
    Scrapes transition and block information from the comments.
    """

    config = RecursiveDict()

    # start parsing from the beginning
    file.seek(0)
    lines = file.readlines()

    # Use simple electron configuration structure for hydrogen
    if element == hydrogen:

        pec_index_header_match = '^C\s*ISEL\s*WAVELENGTH\s*TRANSITION\s*TYPE'
        while not re.match(pec_index_header_match, lines[0], re.IGNORECASE):
            lines.pop(0)
        index_lines = lines

        for i in range(len(index_lines)):

            pec_hydrogen_transition_match = '^C\s*([0-9]*)\.\s*([0-9]*\.[0-9]*)\s*N=\s*([0-9]*) - N=\s*([0-9]*)\s*([A-Z]*)'
            match = re.match(pec_hydrogen_transition_match, index_lines[i], re.IGNORECASE)
            if not match:
                continue

            block_num = int(match.groups()[0])
            wavelength = float(match.groups()[1]) / 10  # convert Angstroms to nm
            upper_level = int(match.groups()[2])
            lower_level = int(match.groups()[3])
            rate_type_adas = match.groups()[4]
            if rate_type_adas == 'EXCIT':
                rate_type = 'excitation'
            elif rate_type_adas == 'RECOM':
                rate_type = 'recombination'
            elif rate_type_adas == 'CHEXC':
                rate_type = 'cx_thermal'
            else:
                raise ValueError("Unrecognised rate type - {}".format(rate_type_adas))

            config[rate_type][element][ionisation][(upper_level, lower_level)] = block_num
            config["wavelength"][element][ionisation][(upper_level, lower_level)] = wavelength

    # Use simple electron configuration structure for hydrogen-like ions
    elif element.atomic_number - ionisation == 1:

        pec_index_header_match = '^C\s*ISEL\s*WAVELENGTH\s*TRANSITION\s*TYPE'
        while not re.match(pec_index_header_match, lines[0], re.IGNORECASE):
            lines.pop(0)
        index_lines = lines

        for i in range(len(index_lines)):

            pec_full_transition_match = '^C\s*([0-9]*)\.\s*([0-9]*\.[0-9]*)\s*([0-9]*)[\(\)\.0-9\s]*-\s*([0-9]*)[\(\)\.0-9\s]*([A-Z]*)'
            match = re.match(pec_full_transition_match, index_lines[i], re.IGNORECASE)
            if not match:
                continue

            block_num = int(match.groups()[0])
            wavelength = float(match.groups()[1]) / 10  # convert Angstroms to nm
            upper_level = int(match.groups()[2])
            lower_level = int(match.groups()[3])
            rate_type_adas = match.groups()[4]
            if rate_type_adas == 'EXCIT':
                rate_type = 'excitation'
            elif rate_type_adas == 'RECOM':
                rate_type = 'recombination'
            elif rate_type_adas == 'CHEXC':
                rate_type = 'cx_thermal'
            else:
                raise ValueError("Unrecognised rate type - {}".format(rate_type_adas))

            config[rate_type][element][ionisation][(upper_level, lower_level)] = block_num
            config["wavelength"][element][ionisation][(upper_level, lower_level)] = wavelength

    # Use full electron configuration structure for anything else
    else:

        configuration_lines = []
        configuration_dict = {}

        configuration_header_match = '^C\s*Configuration\s*\(2S\+1\)L\(w-1/2\)\s*Energy \(cm\*\*-1\)$'
        while not re.match(configuration_header_match, lines[0], re.IGNORECASE):
            lines.pop(0)
        pec_index_header_match = '^C\s*ISEL\s*WAVELENGTH\s*TRANSITION\s*TYPE'
        while not re.match(pec_index_header_match, lines[0], re.IGNORECASE):
            configuration_lines.append(lines[0])
            lines.pop(0)
        index_lines = lines

        for i in range(len(configuration_lines)):

            configuration_string_match = "^C\s*([0-9]*)\s*((?:[0-9][SPDFG][0-9]\s)*)\s*\(([0-9]*\.?[0-9]*)\)([0-9]*)\(\s*([0-9]*\.?[0-9]*)\)"
            match = re.match(configuration_string_match, configuration_lines[i], re.IGNORECASE)
            if not match:
                continue

            config_id = int(match.groups()[0])
            electron_configuration = match.groups()[1].rstrip().lower()
            spin_multiplicity = match.groups()[2]  # (2S+1)
            total_orbital_quantum_number = _L_LOOKUP[int(match.groups()[3])]  # L
            total_angular_momentum_quantum_number = match.groups()[4]  # J

            configuration_dict[config_id] = (electron_configuration + " " + spin_multiplicity +
                                             total_orbital_quantum_number + total_angular_momentum_quantum_number)

        for i in range(len(index_lines)):

            pec_full_transition_match = '^C\s*([0-9]*)\.\s*([0-9]*\.[0-9]*)\s*([0-9]*)[\(\)\.0-9\s]*-\s*([0-9]*)[\(\)\.0-9\s]*([A-Z]*)'
            match = re.match(pec_full_transition_match, index_lines[i], re.IGNORECASE)
            if not match:
                continue

            block_num = int(match.groups()[0])
            wavelength = float(match.groups()[1]) / 10  # convert Angstroms to nm
            upper_level_id = int(match.groups()[2])
            upper_level = configuration_dict[upper_level_id]
            lower_level_id = int(match.groups()[3])
            lower_level = configuration_dict[lower_level_id]
            rate_type_adas = match.groups()[4]
            if rate_type_adas == 'EXCIT':
                rate_type = 'excitation'
            elif rate_type_adas == 'RECOM':
                rate_type = 'recombination'
            elif rate_type_adas == 'CHEXC':
                rate_type = 'cx_thermal'
            else:
                raise ValueError("Unrecognised rate type - {}".format(rate_type_adas))

            config[rate_type][element][ionisation][(upper_level, lower_level)] = block_num
            config["wavelength"][element][ionisation][(upper_level, lower_level)] = wavelength

    return config


def _extract_rate(file, block_num):
    """
    Reads and converts the rate data for the specified block.
    """

    # search from start of file
    file.seek(0)

    wavelength_match = '^\s*[0-9]*\.[0-9] ?A .*$'
    block_id_match = '^\s*[0-9]*\.[0-9] ?A\s*([0-9]*)\s*([0-9]*).*/TYPE = ([a-zA-Z]*).*/ISEL *= * ([0-9]*)$'

    for block in _group_by_block(file, wavelength_match):
        match = re.match(block_id_match, block[0], re.IGNORECASE)

        if not match:
            continue

        if int(match.groups()[3]) == block_num:
            # get number of n, T and rate data points:
            num_n = int(match.groups()[0])
            num_t = int(match.groups()[1])
            num_r = num_n * num_t

            block.pop(0)

            # Load density values
            nn = 0
            density = []
            while nn != num_n:
                next_line = block.pop(0)
                components = next_line.split()
                for value in components:
                    nn += 1
                    density.append(float(value))

            # Load temperature values
            nt = 0
            temperature = []
            while nt != num_t:
                next_line = block.pop(0)
                components = next_line.split()
                for value in components:
                    nt += 1
                    temperature.append(float(value))

            # Load rate values
            nr = 0
            rates = []
            while nr != num_r:
                next_line = block.pop(0)
                components = next_line.split()
                for value in components:
                    nr += 1
                    rates.append(float(value))

            density = np.array(density)
            temperature = np.array(temperature)
            rates = np.array(rates)
            rates = rates.reshape((num_n, num_t))

            # convert units from cm^-3 to m^-3
            density = PerCm3ToPerM3.to(density)
            rates = Cm3ToM3.to(rates)

            return {'ne': density, 'te': temperature, 'rate': rates}

    # If code gets to here, block wasn't found.
    raise RuntimeError('Block number {} was not found in the ADF15 file.'.format(block_num))


def _group_by_block(source_file, match_string):
    """
    Generator the splits the ADF15 file into blocks.

    Groups lines of file into blocks based on precursor '  6561.9A   24...'

    Note: comment section not filtered out of last block, don't over-read!
    """

    buffer = []
    for line in source_file:
        if re.match(match_string, line, re.IGNORECASE):
            if buffer:
                yield buffer
            buffer = [line]
        else:
            buffer.append(line)
    yield buffer



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from cherab.core.utility import RecursiveDict
from .utility import parse_adas2x_rate


def parse_adf22bmp(beam_species, beam_metastable, target_ion, target_ionisation, adf_file_path):
    """
    Opens and parses ADAS ADF22 BMP data files.

    :param beam_species: Element object describing the beam species.
    :param beam_metastable: The metastable level of the beam species.
    :param target_ion: Element object describing the target ion species.
    :param target_ionisation: Ionisation level of the target species.
    :param adf_file_path: Path to ADF15 file from ADAS root.
    :return: Dictionary containing rates.
    """

    rate = RecursiveDict()
    with open(adf_file_path, 'r') as file:
        rate[beam_species][beam_metastable][target_ion][target_ionisation] = parse_adas2x_rate(file)
    return rate


def parse_adf22bme(beam_species, target_ion, target_ionisation, transition, adf_file_path):
    """
    Opens and parses ADAS ADF22 BME data files.

    :param beam_species: Element object describing the beam species.
    :param target_ion: Element object describing the target ion species.
    :param target_ionisation: Ionisation level of the target species.
    :param transition: Atomic transition tuple (upper level, lower level).
    :param adf_file_path: Path to ADF15 file from ADAS root.
    :return: Dictionary containing rates.
    """

    rate = RecursiveDict()
    with open(adf_file_path, 'r') as file:
        rate[beam_species][target_ion][target_ionisation][transition] = parse_adas2x_rate(file)
    return rate



from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize
import sys
import numpy
import os
import os.path as path
import multiprocessing

threads = multiprocessing.cpu_count()
force = False
profile = False
install_rates = False

if "--force" in sys.argv:
    force = True
    del sys.argv[sys.argv.index("--force")]

if "--profile" in sys.argv:
    profile = True
    del sys.argv[sys.argv.index("--profile")]

if "--install-rates" in sys.argv:
    install_rates = True
    del sys.argv[sys.argv.index("--install-rates")]

compilation_includes = [".", numpy.get_include()]
compilation_args = []
cython_directives = {
    'language_level': 3
}

setup_path = path.dirname(path.abspath(__file__))

# build .pyx extension list
extensions = []
for root, dirs, files in os.walk(setup_path):
    for file in files:
        if path.splitext(file)[1] == ".pyx":
            pyx_file = path.relpath(path.join(root, file), setup_path)
            module = path.splitext(pyx_file)[0].replace("/", ".")
            extensions.append(Extension(module, [pyx_file], include_dirs=compilation_includes, extra_compile_args=compilation_args),)

if profile:
    cython_directives["profile"] = True

# generate .c files from .pyx
extensions = cythonize(extensions, nthreads=multiprocessing.cpu_count(), force=force, compiler_directives=cython_directives)

# parse the package version number
with open(path.join(path.dirname(__file__), 'cherab/openadas/VERSION')) as version_file:
    version = version_file.read().strip()

setup(
    name="cherab-openadas",
    version=version,
    license="EUPL 1.1",
    namespace_packages=['cherab'],
    description='Cherab spectroscopy framework: OpenADAS atomic data source package',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics"
    ],
    install_requires=['cherab', 'numpy', 'cython>=0.28'],
    packages=find_packages(),
    include_package_data=True,
    ext_modules=extensions
)

# setup a rate repository with common rates
if install_rates:
    try:
        from cherab.openadas import repository
        repository.populate()
    except ImportError:
        pass




# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
import os
import urllib
from cherab.openadas import repository
from cherab.openadas.parse import *


ADAS_DOWNLOAD_CACHE = os.path.expanduser('~/.cherab/openadas/download_cache')
OPENADAS_FILE_URL = 'http://open.adas.ac.uk/download/'


def install_files(configuration, download=False, repository_path=None, adas_path=None):

    for adf in configuration:
        if adf.lower() == 'adf12':
            for args in configuration[adf]:
                install_adf12(*args, download=download, repository_path=repository_path, adas_path=adas_path)
        if adf.lower() == 'adf15':
            for args in configuration[adf]:
                install_adf15(*args, download=download, repository_path=repository_path, adas_path=adas_path)
        if adf.lower() == 'adf21':
            for args in configuration[adf]:
                install_adf21(*args, download=download, repository_path=repository_path, adas_path=adas_path)
        if adf.lower() == 'adf22bmp':
            for args in configuration[adf]:
                install_adf22bmp(*args, download=download, repository_path=repository_path, adas_path=adas_path)
        if adf.lower() == 'adf22bme':
            for args in configuration[adf]:
                install_adf22bme(*args, download=download, repository_path=repository_path, adas_path=adas_path)


# todo: move print calls to logging
def install_adf12(donor_ion, donor_metastable, receiver_ion, receiver_ionisation, file_path, download=False, repository_path=None, adas_path=None):
    """
    Adds the rates in the ADF12 file to the repository.

    :param donor_ion: The donor ion element described by the rate file.
    :param donor_metastable: The donor ion metastable level.
    :param receiver_ion: The receiver ion element described by the rate file.
    :param receiver_ionisation: The receiver ion ionisation level described by the rate file.
    :param file_path: Path relative to ADAS root.
    :param download: Attempt to download file if not present (Default=True).
    :param repository_path: Path to the repository in which to install the rates (optional).
    :param adas_path: Path to ADAS files repository (optional).
    """

    print('Installing {}...'.format(file_path))
    path = _locate_adas_file(file_path, download, adas_path)
    if not path:
        raise ValueError('Could not locate the specified ADAS file.')

    # decode file and write out rates
    rates = parse_adf12(donor_ion, donor_metastable, receiver_ion, receiver_ionisation, path)
    repository.update_beam_cx_rates(rates, repository_path)


# todo: move print calls to logging
def install_adf15(element, ionisation, file_path, download=False, repository_path=None, adas_path=None):
    """
    Adds the rates in the ADF15 file to the repository.

    :param element: The element described by the rate file.
    :param ionisation: The ionisation level described by the rate file.
    :param file_path: Path relative to ADAS root.
    :param download: Attempt to download file if not present (Default=True).
    :param repository_path: Path to the repository in which to install the rates (optional).
    :param adas_path: Path to ADAS files repository (optional).
    """

    print('Installing {}...'.format(file_path))
    path = _locate_adas_file(file_path, download, adas_path)
    if not path:
        raise ValueError('Could not locate the specified ADAS file.')

    # decode file and write out rates
    rates, wavelengths = parse_adf15(element, ionisation, path)
    repository.update_pec_rates(rates, repository_path)
    repository.update_wavelengths(wavelengths, repository_path)


# todo: move print calls to logging
def install_adf21(beam_species, target_ion, target_ionisation, file_path, download=False, repository_path=None, adas_path=None):
    # """
    # Adds the rate defined in an ADF21 file to the repository.
    #
    # :param file_path: Path relative to ADAS root.
    # :param download: Attempt to download file if not present (Default=True).
    # :param repository_path: Path to the repository in which to install the rates (optional).
    # :param adas_path: Path to ADAS files repository (optional).
    # """

    print('Installing {}...'.format(file_path))
    path = _locate_adas_file(file_path, download, adas_path)
    if not path:
        raise ValueError('Could not locate the specified ADAS file.')

    # # decode file and write out rates
    rate = parse_adf21(beam_species, target_ion, target_ionisation, path)
    repository.update_beam_stopping_rates(rate, repository_path)


# todo: move print calls to logging
def install_adf22bmp(beam_species, beam_metastable, target_ion, target_ionisation, file_path, download=False, repository_path=None, adas_path=None):
    pass
    # """
    # Adds the rate defined in an ADF21 file to the repository.
    #
    # :param file_path: Path relative to ADAS root.
    # :param download: Attempt to download file if not present (Default=True).
    # :param repository_path: Path to the repository in which to install the rates (optional).
    # :param adas_path: Path to ADAS files repository (optional).
    # """

    print('Installing {}...'.format(file_path))
    path = _locate_adas_file(file_path, download, adas_path)
    if not path:
        raise ValueError('Could not locate the specified ADAS file.')

    # # decode file and write out rates
    rate = parse_adf22bmp(beam_species, beam_metastable, target_ion, target_ionisation, path)
    repository.update_beam_population_rates(rate, repository_path)


# todo: move print calls to logging
def install_adf22bme(beam_species, target_ion, target_ionisation, transition, file_path, download=False, repository_path=None, adas_path=None):
    pass
    # """
    # Adds the rate defined in an ADF21 file to the repository.
    #
    # :param file_path: Path relative to ADAS root.
    # :param download: Attempt to download file if not present (Default=True).
    # :param repository_path: Path to the repository in which to install the rates (optional).
    # :param adas_path: Path to ADAS files repository (optional).
    # """

    print('Installing {}...'.format(file_path))
    path = _locate_adas_file(file_path, download, adas_path)
    if not path:
        raise ValueError('Could not locate the specified ADAS file.')

    # # decode file and write out rates
    rate = parse_adf22bme(beam_species, target_ion, target_ionisation, transition, path)
    repository.update_beam_emission_rates(rate, repository_path)


def _locate_adas_file(file_path, download=False, adas_path=None):

    path = None

    # is file in adas path?
    if adas_path:
        target = os.path.join(adas_path, file_path)
        if os.path.isfile(target):
            path = target

    # download file?
    if not path and download:
        target = os.path.join(ADAS_DOWNLOAD_CACHE, file_path)

        # is file in cache? if not download...
        if os.path.isfile(target):
            path = target
        else:

            # create directory structure if missing
            directory = os.path.dirname(target)
            if not os.path.isdir(directory):
                os.makedirs(directory)

            print(" - downloading ADF file '{}' to '{}'".format(file_path, target))

            url = urllib.parse.urljoin(OPENADAS_FILE_URL, file_path.replace('#', '][').lstrip('/'))
            urllib.request.urlretrieve(url, target)
            path = target

    return path



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import numpy as np
from cherab.core.utility.conversion import Cm3ToM3, PerCm3ToPerM3


def parse_adas2x_rate(file):
    """
    Read and parse data from the supplied adf21/22 file stream.

    :param file: A file stream.
    :return: A dictionary.
    """

    raw = {}

    line = file.readline()
    raw['ZT'] = int(line[3:5])                  # target ion charge (int)
    raw['SVREF'] = float(line[13:22])           # coefficient at reference conditions (float)
    raw['SPEC'] = line[29:31]                   # target element (string)
    raw['DATE'] = line[38:46]                   # date of calculation (string)
    raw['CODE'] = line[53:-1]                   # processing code (string)

    file.readline()  # jump the hyphen line

    line = file.readline()
    neb = int(line[1:5])
    ndt = int(line[6:10])
    raw['TREF'] = float(line[17:26])            # reference target temperature (float)

    file.readline()  # jump the hyphen line

    raw['EB'] = readvalues(file, neb, 8)        # beam energy coordinates (1D array)
    raw['DT'] = readvalues(file, ndt, 8)        # target density coordinates (1D array)

    file.readline()  # jump the hyphen line

    sv = np.zeros((neb, ndt))
    for index in range(ndt):
        sv[:, index] = readvalues(file, neb, 8)

    # coefficients at beam energy coordinates (first index) and target density coordinates (second index) (2D array)
    raw['SV'] = sv

    file.readline()  # jump the hyphen line

    line = file.readline()
    ntt = int(line[1:5])
    raw['EREF'] = float(line[12:21])            # reference beam energy (float)
    raw['DREF'] = float(line[28:37])            # reference target density (float)

    file.readline()  # jump the hyphen line

    raw['TT'] = readvalues(file, ntt, 8)        # target temperature coordinates (1D array)

    file.readline()  # jump the hyphen line

    raw['SVT'] = readvalues(file, ntt, 8)       # coefficients at target temperature coordinates (1D array)

    # return essential data and convert units from cm^3 to m^3
    return {
        'e': np.array(raw['EB'], np.float64),
        'n': PerCm3ToPerM3.to(np.array(raw['DT'], np.float64)),
        't': np.array(raw['TT'], np.float64),

        'sen': Cm3ToM3.to(np.array(raw['SV'], np.float64)),
        'st': Cm3ToM3.to(np.array(raw['SVT'], np.float64)),

        'eref': raw['EREF'],
        'nref': PerCm3ToPerM3.to(raw['DREF']),
        'tref': raw['TREF'],
        'sref': Cm3ToM3.to(raw['SVREF'])
    }


def readvalues(file, nb_values, values_per_line, type=float):
    """
    Read and return a given number of values in a file, taking into account
    end of lines. The reading begins at the current read line of the file (which
    must be open to use this function). The read lines of the file are assumed
    to be shaped as following:
    a first useless character, then a given number of 10 characters values, and
    any other characters after (not read).

    :param file: file in which values have to be read
    :param nb_values: number of values to be read
    :param values_per_line: number of values per line on the file
    :param type: python type of the values to be returned
    :return: a numpy 1D array with the read values in the reading order.
    """
    nb_read = 0
    output = []

    while nb_read < nb_values:
        nb_read_line = nb_read % values_per_line
        if nb_read_line == 0:
            line = file.readline()

        output.append(type(line[1+nb_read_line*10:(nb_read_line+1)*10].replace('D', 'E')))
        nb_read += 1

    return np.array(output)


# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import os

"""
Utilities for managing the local rate repository.
"""

DEFAULT_REPOSITORY_PATH = os.path.expanduser('~/.cherab/openadas/repository')


def encode_transition(transition):
    """
    Generate a key string from a transition.

    Both integer and string transition descriptions are handled.
    """

    upper, lower = transition

    upper = str(upper).lower()
    lower = str(lower).lower()

    return '{} -> {}'.format(upper, lower)


def valid_ionisation(element, ionisation):
    """
    Returns true if the element can be ionised to the specified level.

    :param ionisation: Integer ionisation level.
    :return: True/False.
    """
    return ionisation <= element.atomic_number






# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import os
import json
from cherab.core.utility import RecursiveDict
from cherab.core.atomic import Element
from .utility import DEFAULT_REPOSITORY_PATH, valid_ionisation, encode_transition

"""
Utilities for managing the local rate repository - wavelength section.
"""


def add_wavelength(element, ionisation, transition, wavelength, repository_path=None):
    """
    Adds a single wavelength to the repository.

    If adding multiple wavelengths, consider using the update_wavelengths()
    function instead. The update function avoid repeatedly opening and closing
    the rate files.

    :param element:
    :param ionisation:
    :param transition:
    :param wavelength:
    :param repository_path:
    """

    update_wavelengths({
        element: {
            ionisation: {
                transition: wavelength
            }
        }
    }, repository_path)


def update_wavelengths(wavelengths, repository_path=None):

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    for element, ionisations in wavelengths.items():
        for ionisation, transitions in ionisations.items():

            # sanitise and validate
            if not isinstance(element, Element):
                raise TypeError('The element must be an Element object.')

            if not valid_ionisation(element, ionisation):
                raise ValueError('Ionisation level is larger than the number of protons in the element.')

            path = os.path.join(repository_path, 'wavelength/{}/{}.json'.format(element.symbol.lower(), ionisation))

            # read in any existing wavelengths
            try:
                with open(path, 'r') as f:
                    content = RecursiveDict.from_dict(json.load(f))
            except FileNotFoundError:
                content = RecursiveDict()

            # add/replace data for a transition
            for transition in transitions:
                key = encode_transition(transition)
                content[key] = float(wavelengths[element][ionisation][transition])

            # create directory structure if missing
            directory = os.path.dirname(path)
            if not os.path.isdir(directory):
                os.makedirs(directory)

            # write new data
            with open(path, 'w') as f:
                json.dump(content, f, indent=2, sort_keys=True)


def get_wavelength(element, ionisation, transition, repository_path=None):

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH
    path = os.path.join(repository_path, 'wavelength/{}/{}.json'.format(element.symbol.lower(), ionisation))
    try:
        with open(path, 'r') as f:
            content = json.load(f)
        return content[encode_transition(transition)]
    except (FileNotFoundError, KeyError):
        raise RuntimeError('Requested wavelength (element={}, ionisation={}, transition={})'
                           ' is not available.'.format(element.symbol, ionisation, transition))

